# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
All problems in computer science can be solved by another layer of
indirection, except for the problem of too many layers of indirection.


Traits overview
---------------

Traits classes are separated into two modules:

core:
    TraitsInfo          meta-data container for traits system
    MetaType            base class for traited class creation
    Type                base traited class 

mapped:
    MappedType             basic class for traited class mapped to db
      * (Type)             traits mapped to columns
      * (MappedType)       traits mapped to column(foreignkey) -> table


Traits metadata
---------------

While the traits of a class can be declared nearly arbitrarily, there are
intrinsic pieces of information in the meta-data required to make the traits
system work:

        doc         string      long description, possibly longer string
        label       string      short name appearing in UI
        default     object      default value of attribute
        required    boolean     determines whether must be set before storage
        range       Range       helps validate or specify parameter variation

.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: marmaduke <duke@eml.cc>

"""

import abc
import six
from copy import deepcopy, copy
from tvb.basic.traits.util import get, Args, TypeRegister, ispublic
from tvb.basic.logger.builder import get_logger


LOG = get_logger(__name__)

KWARG_SELECT_MULTIPLE = 'select_multiple'
KWARG_ORDER = 'order'                   # -1 value means hidden from UI
KWARG_AVOID_SUBCLASSES = 'fixed_type'   # When set on a traited attr, no subclasses will be returned
KWARG_FILE_STORAGE = 'file_storage'
KWARG_REQUIRED = 'required'
KWARG_FILTERS_UI = 'filters_ui'
KWARG_FILTERS_BACKEND = 'filters_backend'
KWARG_OPTIONS = 'options'               # Used for Enumerate basic type
KWARG_STORAGE_PATH = 'storage_path'
KWARS_USE_STORAGE = 'use_storage'
KWARS_STORED_METADATA = 'stored_metadata'
KWARGS_LOAD_DEFAULT = "load_default"

FILE_STORAGE_DEFAULT = 'HDF5'
FILE_STORAGE_EXPAND = 'expandable_HDF5'
FILE_STORAGE_NONE = 'None'

SPECIAL_KWDS = ['bind',                 # set and used internally by the traiting mechanism
                'doc', 'label', KWARG_REQUIRED, 'locked', 'default', 'range',
                'configurable_noise', KWARG_OPTIONS, KWARS_STORED_METADATA,
                KWARG_AVOID_SUBCLASSES, KWARG_FILTERS_UI, KWARG_FILTERS_BACKEND, KWARG_SELECT_MULTIPLE, KWARG_ORDER,
                KWARG_FILE_STORAGE, KWARS_USE_STORAGE, KWARGS_LOAD_DEFAULT]


## Module global used by MetaType.
TYPE_REGISTER = TypeRegister()



class TraitsInfo(dict):
    """
    TraitsInfo is a container for information related to the owner
    class and its traited attributes that is used by the traits
    classes. It is needed because many of the attribute names used
    by traits, e.g. data & bound, may mean other things to other
    classes and we can't step on their toes.

    TraitsInfo is a dictionary of the owner's traited attributes, and its other attributes are:

    - name - name of attribute on owner class
    - bound - whether the trait is bound as data descriptor
    - wraps - class this trait wraps
    - inits - a namedtuple of positional and keyword arguments
              given to intialize the trait instance
    - value - instance value of trait, equal to trait if wraps==None
    - defaults - default args to wraps' constructor

    """


    def __init__(self, trait, name='<no name!>', bound=False, wraps=None,
                 inits=Args((), {}), value=None, wraps_defaults=()):
        self.name = name
        self.bound = bound
        self.wraps = wraps
        self.wraps_defaults = wraps_defaults
        self.inits = inits
        self.value = value


    @property
    def file_storage(self):
        if KWARG_FILE_STORAGE not in self.inits.kwd:
            return FILE_STORAGE_DEFAULT
        return self.inits.kwd[KWARG_FILE_STORAGE]


    @property
    def order_number(self):
        if KWARG_ORDER not in self.inits.kwd:
            return 0
        return self.inits.kwd[KWARG_ORDER]


    @property
    def required(self):
        if KWARG_REQUIRED not in self.inits.kwd:
            return True
        return self.inits.kwd[KWARG_REQUIRED]


    @property
    def use_storage(self):
        if KWARS_USE_STORAGE not in self.inits.kwd:
            return True
        return self.inits.kwd[KWARS_USE_STORAGE]


    @property
    def stored_metadata(self):
        if KWARS_STORED_METADATA not in self.inits.kwd:
            return None
        return self.inits.kwd[KWARS_STORED_METADATA]


    @property
    def range_interval(self):
        if 'range' not in self.inits.kwd:
            return None
        return self.inits.kwd['range']


    @property
    def select_multiple(self):
        if KWARG_SELECT_MULTIPLE in self.inits.kwd:
            return self.inits.kwd[KWARG_SELECT_MULTIPLE]
        return False


    def __repr__(self):
        return 'TraitsInfo(%r)' % (super(TraitsInfo, self).__repr__(), )


    def copy(self):
        """
        Create a copy for current Traits.
        """
        new_value = deepcopy(self.value)
        copyed = TraitsInfo(new_value, self.name, self.bound,
                            self.wraps, self.inits, wraps_defaults=self.wraps_defaults)
        for key, value in six.iteritems(self):
            copyed[key] = copy(value)
        return copyed



class Type(object):
    """ Type is redefined below. This is here so MetaType can refer to Type """



class MetaType(abc.ABCMeta):
    """

    The MetaType class augments the class creation and instantiation of all the
    types in the Traits system. See the doc-strings of the methods for more
    details:

        __new__  - creates a class 
        __call__ - create a class instance

    While the basic Traits mechanisms are described and implemented in this
    class, see DeclarativeMetaType for implementation and description of 
    database mapping of Traits classes.

    """


    def __new__(mcs, name, bases, dikt):
        """
        MetaType.__new__ creates a new class, but is typically involved
        *implicitly* either by declaring the __metaclass__ attribute

            class Type(object):
                __metaclass__ = MetaType

        or by subclassing Type:

            class Foo(Type):
                pass

        but in both cases, this method is called to produce the new class object.

        To setup a trait attribute, we :
            - check if it's a type, if so, instantiate
            - tell the attr it's name on owner class 
            - setup private attr with attr.value 
            - augment owner class doc-string with trait description 
            - add trait to information on class 
        """

        # if we're wrapping a class, pop that out
        wraps = dikt.pop('wraps', set([]))
        wraps_defaults = dikt.pop('defaults', ())

        # make new class
        newcls = super(MetaType, mcs).__new__(mcs, name, bases, dikt)
        # add new class to type register
        TYPE_REGISTER.append(newcls)

        # prep class doc string
        doc = get(dikt, '__doc__', 'traited class ' + name)
        doc += "\n    **traits on this class:**\n"

        # build traits info on newcls' attrs
        if hasattr(newcls, 'trait'):
            trait = newcls.trait.copy()
        else:
            trait = TraitsInfo(newcls)

        for key in list(filter(ispublic, dir(newcls))):
            attr = getattr(newcls, key)
            if isinstance(attr, MetaType) or isinstance(attr, Type):
                if isinstance(attr, MetaType):
                    attr = attr()
                try:
                    attr = deepcopy(attr)
                except Exception:
                    attr = copy(attr)
                attr.trait.name = key
                setattr(newcls, key, attr)
                doc += "\n\t``%s`` (%s)\n" % (key, str(attr.trait.inits.kwd.get('label', "")))
                doc += "\t\t| %s\n" % str(attr.trait.inits.kwd.get('doc', "")).replace("\n", " ")
                doc += "\t\t| ``default``:  %s \n" % str(attr.trait.inits.kwd.get('default', None)).replace("\n", " ")
                specified_range = attr.trait.inits.kwd.get('range', None)
                if specified_range:
                    doc += "\t\t| ``range``: low = %s ; high = %s \n\t\t\n" % (str(specified_range.lo),
                                                                               str(specified_range.hi))
                trait[key] = attr

        # add info to new class
        if wraps:
            trait.wraps = wraps
        if wraps_defaults:
            trait.wraps_defaults = wraps_defaults
        newcls.trait = trait

        # bind traits unless told otherwise
        for attr in newcls.trait.values():
            attr.trait.bound = attr.trait.inits.kwd.get('bind', True)

        newcls.__doc__ = doc
        return newcls


    def __call__(cls, *args, **kwds):
        """
        MetaType.__call__ method wraps ncs.__init__(ncs.__new__(*, **), *, **),
        and is implicitly called when the class __init__()s.

        b.Range(*args, **kwds) ->
        b.Range.__init__(b.Range.__new__(MetaType.__call__(b.Range, *args, **kwds), *, **), *, **)

        When creating instances of Traits classes, we

            - if wrapping, try to instantiation wrapped class 
            - check keyword arguments, use to initialize trait attributes
            - record all other keyword args for later use 
            - create class instance 
            - return instance updated with information
        """

        inits = Args(args, kwds.copy())
        if 'default' in kwds:
            value = kwds.pop('default')
            if isinstance(value, MetaType):
                value = value()

        elif cls.trait.wraps:
            wrapped_callable = cls.trait.wraps[0] if isinstance(cls.trait.wraps, tuple) else cls.trait.wraps
            # no args, and we have defaults
            if cls.trait.wraps_defaults:
                _args, _kwds = cls.trait.wraps_defaults
                value = wrapped_callable(*_args, **_kwds)
            # else default constructor, no args
            else:
                value = wrapped_callable()
        else:
            value = None

        kwdtraits = {}
        for key in set(kwds.keys()) & set(cls.trait.keys()):
            kwdtraits[key] = kwds[key]
            del kwds[key]

        options = kwds.get(KWARG_OPTIONS, None)

        # discard kwds to be passed for instantiation
        [kwds.pop(key, None) for key in SPECIAL_KWDS]

        ## 1. - Call __init__ to get a new instance (includes initialization of SqlAlchemy fields)
        try:
            inst = super(MetaType, cls).__call__(*args, **kwds)
        except Exception:
            if not args and not kwds:
                raise
            else:
                # most likely case is that a keyword argument is misspelled.
                msg = "Couldn't create instance of %s with args: %r, %r."
                msg %= cls, args, kwds
                LOG.exception(msg)
                raise TypeError(msg)

        ## 2. - Populate default fields from Trait arguments:
        inst.trait = cls.trait.copy()
        inst.trait.options = options
        inst.trait.value = deepcopy(value)
        inst.trait.inits = inits

        for name, attr in six.iteritems(inst.trait):
            try:
                setattr(inst, name, deepcopy(attr.trait.value))
            except Exception:
                LOG.exception("Could not set attribute '" + name + "' on " + str(inst.__class__.__name__))
                raise

        ## 3. - Load console defaults, in case requested:
        if KWARGS_LOAD_DEFAULT in inits.kwd and inits.kwd[KWARGS_LOAD_DEFAULT]:
            cls.from_file(instance=inst)

        ## 4. - Overwrite with **kwargs from constructor call:
        for name, attr in six.iteritems(kwdtraits):
            try:
                setattr(inst, name, attr)
            except Exception:
                LOG.exception("Could not set kw-given attribute '" + name + "' on " + str(inst.__class__.__name__))
                raise

        # The owner class, if any, will set this to true, see metatype.__new__
        inst.trait.bound = False

        LOG.debug('%s initialized', inst)

        return inst



class Type(object):
    """
    Type class provides a base class for dataTypes and the attributes on dataTypes.

    When a Type instance is an attribute of a class and self.bound is True, the
    instance will act as a data descriptor, setting/getting its corresponding
    value on the owner class.

    In the case of sql'ed values, names are coordinated such that the private
    value (```obj._name```) of the public attr (```obj.name```) on the owner
    class used by the Type instance is actually the corresponding sqlAlchemy
    data descriptor as generated by the value of 'sql' keyword to the Type
    instance __init__.
    """

    __metaclass__ = MetaType
    _summary_info = None

    def __get__(self, inst, cls):
        """
        When an attribute of Type is retrieved on another class.
        """
        if inst is None:
            return self
        if self.trait.bound:
            if hasattr(inst, '_' + self.trait.name):
                ## Return simple DB field or cached value
                return getattr(inst, '_' + self.trait.name)
            else:
                return None
        else:
            return self


    def __set__(self, inst, value):
        """
        When an attribute of Type is set on another class
        """
        if self.trait.bound:
            ## First validate that the given value is compatible with the current attribute definition
            accepted_types = [type(self), type(None)]
            if isinstance(self.trait.wraps, tuple):
                accepted_types.extend(self.trait.wraps)
            else:
                accepted_types.append(self.trait.wraps)

            if (type(value) in accepted_types or isinstance(value, type(self))
                or (isinstance(value, (list, tuple)) and self.trait.select_multiple)):
                self._put_value_on_instance(inst, value)
            else:
                msg = 'expected type %s, received type %s' % (type(self), type(value))
                LOG.error(msg)
                raise AttributeError(msg)


    def _put_value_on_instance(self, inst, value):
        """
        Is the ultimate method called by __set__ implementations.
        We write it separately here, because subclasses might need to call this separately, 
        without the __set__ default value validation.
        """
        setattr(inst, '_' + self.trait.name, value)
        inst.trait[self.trait.name].value = value


    def __repr__(self):
        """
        Type.repr builds a useful representation of itself, which can be 
        configured with values in config:
        """
        trait = self.trait
        rep = self.__class__.__name__ + "("
        objstr = object.__repr__(self)
        value = objstr if self is trait.value else trait.value

        reprinfo = {'value': repr(value),
                    'bound': repr(trait.bound)}
        if trait.wraps:
            reprinfo['wraps'] = repr(trait.wraps)
        if trait.bound:
            reprinfo['name'] = repr(trait.name)

        reprstr = ['%s=%s' % (k, v) for k, v in reprinfo.items()]
        return rep + ', '.join(reprstr) + ')'


    def configure(self):
        """
        Call this method to process linked attributes on DataType.
        This will be called before storing entity in DB.
        """
        pass


    @property
    def summary_info(self):
        """
        For a particular DataType, return a dictionary of label: value, 
        to describe the entity from scientific point of view.
        """
        if self._summary_info is None and hasattr(self, "_find_summary_info"):
            self._summary_info = self._find_summary_info()
        return self._summary_info

    # Used by IPython Notebook
    def _repr_html_(self):
        "Generate HTML repr for use in IPython notebook."
        info = self.summary_info
        if info is None or len(info) == 0:
            info = {key: getattr(self, key) for key in self.trait.keys()}
        html = ['<table width=100%>']
        row_fmt = '<tr><td>%s</td><td>%s</td></tr>'
        for key, value in info.items():
            html.append(row_fmt % (key, value))
        return ''.join(html)

    def _find_summary_info(self):
        """
        To be implemented in every subclass.
        """
        return None


TypeBase = Type



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""

The Array datatypes. This brings together the scientific and framework 
methods that are associated with the Array datatypes.

.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>

"""

import numpy
from tvb.basic.traits import core, types_basic as basic
from tvb.basic.traits.types_mapped import MappedType, Array
from tvb.basic.traits.exceptions import ValidationException


class BaseArray(Array):
    "Base class for array-type traits."
    def _find_summary_info(self):
        "Summarize array contents."
        summary = {"Array type": self.__class__.__name__,
                   "Shape": self.shape,
                   "Maximum": self.value.max(),
                   "Minimum": self.value.min(),
                   "Mean": self.value.mean(),
                   "Median": numpy.median(self.value)}
        return summary


class FloatArray(BaseArray):
    _ui_name = "Floating-point array"
    dtype = basic.DType(default=numpy.float64)


class IntegerArray(BaseArray):
    _ui_name = "Array of integers"
    dtype = basic.DType(default=numpy.int32)


class ComplexArray(BaseArray):
    _ui_name = "Array of complex numbers"
    dtype = basic.DType(default=numpy.complex128)
    stored_metadata = [key for key in MappedType.DEFAULT_STORED_ARRAY_METADATA if key != MappedType.METADATA_ARRAY_VAR]


class BoolArray(BaseArray):
    _ui_name = "Boolean array"
    dtype = basic.DType(default=numpy.bool)
    stored_metadata = [MappedType.METADATA_ARRAY_SHAPE]

    def _find_summary_info(self):
        summary = {"Array type": self.__class__.__name__,
                   "Shape": self.shape,
                   'Number True': self.value.sum(),
                   'Percent True': self.value.mean()*100,}
        return summary


class StringArray(BaseArray):
    _ui_name = "Array of strings"
    # if you want variable length strings, you must use dtype=object
    # otherwise, must specify max lenth as 'na' where n is integer,
    # e.g. dtype='100a' for a string w/ max len 100 characters.
    dtype = None
    stored_metadata = [MappedType.METADATA_ARRAY_SHAPE]
    def _find_summary_info(self):
        summary = {"Array type": self.__class__.__name__,
                   "Shape": self.shape}
        return summary


class PositionArray(FloatArray):
    _ui_name = "Array of positions"

    coordinate_system = basic.String(label="Coordinate system",
                                     default="cartesian",
                                     doc="""The coordinate system used to specify the positions.
                                     Eg: 'spherical', 'polar'""")

    coordinate_space = basic.String(label="Coordinate space",
                                    default="None",
                                    doc="The standard space the positions are in, eg, 'MNI', 'colin27'")


class OrientationArray(FloatArray):
    _ui_name = "Array of orientations"
    coordinate_system_or = basic.String(label="Coordinate system",
                                        default="cartesian")


class IndexArray(BaseArray):
    _ui_name = "Index array"
    target = Array(label="Indexed array",
                   file_storage=core.FILE_STORAGE_NONE,
                   doc="A link to the array that the indices index.")


class MappedArray(MappedType):
    "An array stored in the database."
    KEY_SIZE = "size"
    KEY_OPERATION = "operation"

    title = basic.String
    label_x, label_y = basic.String, basic.String
    aggregation_functions = basic.JSONType(required=False)
    dimensions_labels = basic.JSONType(required=False)
    nr_dimensions, length_1d, length_2d, length_3d, length_4d = [basic.Integer] * 5
    array_data = Array()

    __generate_table__ = True

    def _find_summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        summary = {"Title:": self.title,
                   "Dimensions:": self.dimensions_labels}
        summary.update(self.get_info_about_array('array_data'))
        return summary

    @property
    def display_name(self):
        """
        Overwrite from superclass and add title field
        """
        previous = super(MappedArray, self).display_name
        if previous is None:
            return str(self.title)
        return str(self.title) + " " + previous

    @property
    def shape(self):
        """
        Shape for current wrapped NumPy array.
        """
        return self.array_data.shape

    def configure_chunk_safe(self):
        """ Configure part which is chunk safe"""
        data_shape = self.get_data_shape('array_data')
        self.nr_dimensions = len(data_shape)
        for i in range(min(self.nr_dimensions, 4)):
            setattr(self, 'length_%dd' % (i + 1), int(data_shape[i]))

    def configure(self):
        """After populating few fields, compute the rest of the fields"""
        super(MappedArray, self).configure()
        if not isinstance(self.array_data, numpy.ndarray):
            return
        self.nr_dimensions = len(self.array_data.shape)
        for i in range(min(self.nr_dimensions, 4)):
            setattr(self, 'length_%dd' % (i + 1), self.array_data.shape[i])

    @staticmethod
    def accepted_filters():
        filters = MappedType.accepted_filters()
        filters.update({'datatype_class._nr_dimensions': {'type': 'int', 'display': 'Dimensionality',
                                                          'operations': ['==', '<', '>']},
                        'datatype_class._length_1d': {'type': 'int', 'display': 'Shape 1',
                                                      'operations': ['==', '<', '>']},
                        'datatype_class._length_2d': {'type': 'int', 'display': 'Shape 2',
                                                      'operations': ['==', '<', '>']}})
        return filters

    def reduce_dimension(self, ui_selected_items):
        """
        ui_selected_items is a dictionary which contains items of form:
        '$name_$D : [$gid_$D_$I,..., func_$FUNC]' where '$D - int dimension', '$gid - a data type gid',
        '$I - index in that dimension' and '$FUNC - an aggregation function'.

        If the user didn't select anything from a certain dimension then it means that the entire
        dimension is selected
        """

        #The fields 'aggregation_functions' and 'dimensions' will be of form:
        #- aggregation_functions = {dimension: agg_func, ...} e.g.: {0: sum, 1: average, 2: none, ...}
        #- dimensions = {dimension: [list_of_indexes], ...} e.g.: {0: [0,1], 1: [5,500],...}
        dimensions, aggregation_functions, required_dimension, shape_restrictions = \
            self.parse_selected_items(ui_selected_items)

        if required_dimension is not None:
            #find the dimension of the resulted array
            dim = len(self.shape)
            for key in aggregation_functions.keys():
                if aggregation_functions[key] != "none":
                    dim -= 1
            for key in dimensions.keys():
                if (len(dimensions[key]) == 1 and
                    (key not in aggregation_functions
                     or aggregation_functions[key] == "none")):
                    dim -= 1
            if dim != required_dimension:
                self.logger.debug("Dimension for selected array is incorrect")
                raise ValidationException("Dimension for selected array is incorrect!")

        result = self.array_data
        full = slice(0, None)
        cut_dimensions = 0
        for i in range(len(self.shape)):
            if i in dimensions.keys():
                my_slice = [full for _ in range(i - cut_dimensions)]
                if len(dimensions[i]) == 1:
                    my_slice.extend(dimensions[i])
                    cut_dimensions += 1
                else:
                    my_slice.append(dimensions[i])
                result = result[tuple(my_slice)]
            if i in aggregation_functions.keys():
                if aggregation_functions[i] != "none":
                    result = eval("numpy." + aggregation_functions[i] + "(result,axis=" + str(i - cut_dimensions) + ")")
                    cut_dimensions += 1

        #check that the shape for the resulted array respects given conditions
        result_shape = result.shape
        for i in range(len(result_shape)):
            if i in shape_restrictions:
                flag = eval(str(result_shape[i]) + shape_restrictions[i][self.KEY_OPERATION] +
                            str(shape_restrictions[i][self.KEY_SIZE]))
                if not flag:
                    msg = ("The condition is not fulfilled: dimension "
                           + str(i + 1) + " "
                           + shape_restrictions[i][self.KEY_OPERATION] + " "
                           + str(shape_restrictions[i][self.KEY_SIZE]) +
                           ". The actual size of dimension " + str(i + 1)
                           + " is " + str(result_shape[i]) + ".")
                    self.logger.debug(msg)
                    raise ValidationException(msg)

        if required_dimension is not None and 1 <= required_dimension != len(result.shape):
            self.logger.debug("Dimensions of the selected array are incorrect")
            raise ValidationException("Dimensions of the selected array are incorrect!")

        return result

    def parse_selected_items(self, ui_selected_items):
        """
        Used for parsing the user selected items.

        ui_selected_items is a dictionary which contains items of form:
        'name_D : [gid_D_I,..., func_FUNC]' where 'D - dimension', 'gid - a data type gid',
        'I - index in that dimension' and 'FUNC - an aggregation function'.
        """
        expected_shape_str = ''
        operations_str = ''
        dimensions = dict()
        aggregation_functions = dict()
        required_dimension = None
        for key in ui_selected_items.keys():
            split_array = str(key).split("_")
            current_dim = split_array[len(split_array) - 1]
            list_values = ui_selected_items[key]
            if list_values is None or len(list_values) == 0:
                list_values = []
            elif not isinstance(list_values, list):
                list_values = [list_values]
            for item in list_values:
                if str(item).startswith("expected_shape_"):
                    expected_shape_str = str(item).split("expected_shape_")[1]
                elif str(item).startswith("operations_"):
                    operations_str = str(item).split("operations_")[1]
                elif str(item).startswith("requiredDim_"):
                    required_dimension = int(str(item).split("requiredDim_")[1])
                elif str(item).startswith("func_"):
                    agg_func = str(item).split("func_")[1]
                    aggregation_functions[int(current_dim)] = agg_func
                else:
                    str_array = str(item).split("_")
                    if int(str_array[1]) in dimensions:
                        dimensions[int(str_array[1])].append(int(str_array[2]))
                    else:
                        dimensions[int(str_array[1])] = [int(str_array[2])]
        return dimensions, aggregation_functions, required_dimension, self._parse_expected_shape(expected_shape_str,
                                                                                                 operations_str)

    def _parse_expected_shape(self, expected_shape_str='', operations_str=''):
        """
        If we have the inputs of form: expected_shape='x,512,x' and operations='x,&lt;,x'.
        The result will be: {1: {'size':512, 'operation':'<'}}
        """
        result = {}
        if len(expected_shape_str.strip()) == 0 or \
           len(operations_str.strip()) == 0:
            return result

        shape_array = str(expected_shape_str).split(",")
        op_array = str(operations_str).split(",")

        operations = self._get_operations()
        for i in range(len(shape_array)):
            if str(shape_array[i]).isdigit() and i < len(op_array) and op_array[i] in operations:
                result[i] = {self.KEY_SIZE: int(shape_array[i]),
                             self.KEY_OPERATION: operations[op_array[i]]}
        return result

    def read_data_shape(self):
        """ Expose shape read on field 'array_data' """
        return self.get_data_shape('array_data')

    def read_data_slice(self, data_slice):
        """ Expose chunked-data access. """
        return self.get_data('array_data', data_slice)

    @staticmethod
    def _get_operations():
        """Return accepted operations"""
        operations = {'&lt;': '<',
                      '&gt;': '>',
                      '&ge;': '>=',
                      '&le;': '<=',
                      '==': '=='}
        return operations




# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
"""
if __name__ == "__main__":
    from tvb.tests.library import setup_test_console_env
    setup_test_console_env()
    
import unittest
from tvb.datatypes import arrays
from tvb.basic.traits.parameters_factory import get_traited_instance_for_name, get_traited_subclasses
from tvb.basic.traits.types_mapped import Array
from tvb.tests.library.base_testcase import BaseTestCase


class ParametersFactoryTest(BaseTestCase):
    
    def test_traitedsubclassed(self):
        """
        Tests successful creation of traited classes.
        """
        # We imported array so we should have all these traited classes registered
        expected = [
                    'IntegerArray',
                    'StringArray', 'PositionArray',
                    'IndexArray',

                    'BoolArray', 'OrientationArray',


                    'FloatArray',
                    'ComplexArray', 'Array', 'SparseMatrix']
        subclasses = get_traited_subclasses(Array)
        for key in expected:
            self.assertTrue(key in subclasses)
            
            
    def test_get_traited_instance(self):
        """
        Try to create an instance of a class using the traited method.
        """
        inst = get_traited_instance_for_name("StringArray", Array, {})
        self.assertTrue(isinstance(inst, arrays.StringArray))



def suite():
    """
    Gather all the tests in a test suite.
    """
    test_suite = unittest.TestSuite()
    test_suite.addTest(unittest.makeSuite(ParametersFactoryTest))
    return test_suite


if __name__ == "__main__":
    #So you can run tests from this package individually.
    TEST_RUNNER = unittest.TextTestRunner()
    TEST_SUITE = suite()
    TEST_RUNNER.run(TEST_SUITE) 


# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Traits metadata module
======================

introduction
------------

The problem is that various classes have members whose properties we want to
annotate in a way that does not affect the execution of the class algorithms
but that exposes an API to pull out this info for a user interface and
documentation. It should be readable for the programmers as well. We refer to
an annotated/propertied member as a trait.

Consider the following class:

    >>> class Point(object):
    ...     x = 0.0
    ...     y = 0.0

If you want to present this data to the user, you may want to label each
attribute of the Point with more than just 'x' or 'y', or you want to check
that if the user creates a Point, the values have the right type (here, we
assume the type of 0.0 means floating point), and if you want to store the
instances of Point in a data base, each one needs a unique id. These pieces of
information need to be put somewhere, and the best place to put it is in the
attribute declaration, for example,

    >>> class Point(Type):
    ...     x = Float(0.0, doc='horizontal position')
    ...     y = Float(0.0, doc='vertical position')

Here, the type of each attribute is explicit, and a documentation string
is provided, *all in the same attribute declaration*, and this co-location
of information makes it easier to maintain, and elsewhere in the program,
this information can be accessed programmatically to generate user interfaces,
serialize data, validate data, etc. Additionally, all classes' docstrings
will have Traits information automatically added:

    >>> print Point.__doc__
    Traited class Point
        x: horizontal position, type float
        y: vertical position, type float

The Traits package allows us to write classes like the one above by using a
metaclass to customize the class creation. In brief, when Python reads a class
definition the Trait metaclass has access to all the attribute declarations,
and it will place a dictionary of all the data provided in the declarations on
the class.  When the classes are instantiated and used, the Trait attributes
behave exactly like their values, such that class methods do not need to know
they are accessing Trait attributes. For more details, see `traits.core`.


package modules
---------------

    core            Provides the core of Traits, a metaclass and class that
                    implement the reflection/annotation that Traits allows.

    types_basic     Provides basic predefined types as well as the Type that
                    should be inherited from when making new Types

    traited_interface  Provides recursive interfaces classes that are data
                    descriptors, i.e. when placed on classes, they compute
                    the full description of the class and it's Traited
                    attributes.

    util            Utility functions and classes.


automatic class documentation
`````````````````````````````

Classes that subclass Type (or subclasses of MetaType) have their docstrings
automatically augmented by the repr() of each attribute which is a Type. This
allows one to find the trait documentation both in SPhinx and at the console.

References
----------

Suggested reading is `Unifying Types and Classes`_ and th `Python data model`_.

.. _Unifying Types and Classes:
    http://www.python.org/download/releases/2.2.3/descrintro/
.. _Python data model: http://docs.python.org/reference/datamodel.html

.. moduleauthor:: Marmaduke Woodman <mw@eml.cc>

"""

import core
import traited_interface, traited_interface2
from tvb.basic.profile import TvbProfile

# Add interfaces based on configured parameter on classes
setattr(core.Type, TvbProfile.current.TRAITS_CONFIGURATION.interface_method_name,
        traited_interface.TraitedInterfaceGenerator())
setattr(core.Type, 'interface_experimental',
        traited_interface2.TraitedInterfaceGeneratorExperimental())






# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
The Volume datatypes. This brings together the scientific and framework 
methods that are associated with the volume datatypes.

"""

import numpy
from tvb.basic.logger.builder import get_logger
from tvb.basic.traits import types_basic as basic
from tvb.datatypes import volumes, arrays
from tvb.basic.arguments_serialisation import preprocess_space_parameters


LOG = get_logger(__name__)


class VolumetricDataMixin(object):
    """Provides subclasses with useful methods for volumes."""

    def write_data_slice(self, data):
        """
        We are using here the same signature as in TS, just to allow easier parsing code.
        This is not a chunked write.

        """
        self.store_data("array_data", data)

    def get_volume_slice(self, x_plane, y_plane, z_plane):
        slices = slice(self.length_1d), slice(self.length_2d), slice(z_plane, z_plane + 1)
        slice_x = self.read_data_slice(slices)[:, :, 0]  # 2D
        slice_x = numpy.array(slice_x, dtype=int)

        slices = slice(x_plane, x_plane + 1), slice(self.length_2d), slice(self.length_3d)
        slice_y = self.read_data_slice(slices)[0, :, :][..., ::-1]
        slice_y = numpy.array(slice_y, dtype=int)

        slices = slice(self.length_1d), slice(y_plane, y_plane + 1), slice(self.length_3d)
        slice_z = self.read_data_slice(slices)[:, 0, :][..., ::-1]
        slice_z = numpy.array(slice_z, dtype=int)

        return [slice_x, slice_y, slice_z]

    def get_volume_view(self, x_plane, y_plane, z_plane, **kwargs):
        # Work with space inside Volume:
        x_plane, y_plane, z_plane = preprocess_space_parameters(x_plane, y_plane, z_plane, self.length_1d,
                                                                self.length_2d, self.length_3d)
        slice_x, slice_y, slice_z = self.get_volume_slice(x_plane, y_plane, z_plane)
        return [[slice_x.tolist()], [slice_y.tolist()], [slice_z.tolist()]]

    def get_min_max_values(self):
        """
        Retrieve the minimum and maximum values from the metadata.
        :returns: (minimum_value, maximum_value)
        """
        metadata = self.get_metadata('array_data')
        return metadata[self.METADATA_ARRAY_MIN], metadata[self.METADATA_ARRAY_MAX]


class StructuralMRI(VolumetricDataMixin, arrays.MappedArray):
    """
    Quantitative volumetric data recorded by means of Magnetic Resonance Imaging.

    """
    # without the field below weighting and volume columns are going to be added to the MAPPED_ARRAY table
    __generate_table__ = True

    array_data = arrays.FloatArray(label="contrast")

    weighting = basic.String(label="MRI weighting")  # eg, "T1", "T2", "T2*", "PD", ...

    volume = volumes.Volume



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
"""
.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: bogdan.neacsa <bogdan.neacsa@codemart.ro>
"""
if __name__ == "__main__":
    from tvb.tests.library import setup_test_console_env
    setup_test_console_env()
    
import unittest
from tvb.basic.filters.chain import FilterChain
from tvb.basic.filters.exceptions import InvalidFilterChainInput, InvalidFilterEntity
from tvb.tests.library.base_testcase import BaseTestCase


class FilteringTest(BaseTestCase):
    """
    Test that defining and evaluating a filter on entities is correctly processed.
    """
    
    class DummyFilterClass():
        """
        This class is a class with some attributes that is used to test the filtering module.
        """
        attribute_1 = None
        attribute_2 = None
        attribute_3 = None
        
        def __init__(self, attribute_1 = None, attribute_2 = None, attribute_3 = None):
            self.attribute_1 = attribute_1
            self.attribute_2 = attribute_2
            self.attribute_3 = attribute_3
            
        def __str__(self):
            return self.__class__.__name__ + '(attribute_1=%s, attribute_2=%s, attribute_3=%s)'% (
                    self.attribute_1, self.attribute_2, self.attribute_3)
            
            
    def __should_pass(self, filter_entity, dt_entity):
        """Private method for throwing exception when filter does not passes."""
        self.assertTrue(filter_entity.get_python_filter_equivalent(datatype_to_check=dt_entity), 
                        "Filter %s should pass %s but fails it for some reason!"%(filter_entity, dt_entity))
        
        
    def __should_fail(self, filter_entity, dt_entity):
        """Private method for throwing exception when filter does not fail to remove entity."""
        self.assertFalse(filter_entity.get_python_filter_equivalent(datatype_to_check=dt_entity),
                         "Filter %s should drop from results %s but passes it!"%(filter_entity, dt_entity))

    
    def test_input_passed_simple(self):
        """
        Supported operators should be "==", "!=", "<", ">", "in", "not in" so far.
        """
        test_filter = FilterChain(fields = [FilterChain.datatype + '.attribute_1'], 
                                  operations = ["=="], values = ['test_val'])
        self.__should_pass(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 'test_val'))
        self.__should_fail(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 1))
        
        test_filter = FilterChain(fields = [FilterChain.datatype + '.attribute_1'], 
                                  operations = ["!="], values = ['test_val'])
        self.__should_pass(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 'test_val_other'))
        self.__should_fail(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 'test_val'))
        
        test_filter = FilterChain(fields = [FilterChain.datatype + '.attribute_1'], 
                                  operations = [">"],  values = [3])
        self.__should_pass(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 5))
        self.__should_fail(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 1))
        
        test_filter = FilterChain(fields = [FilterChain.datatype + '.attribute_1'], 
                                  operations = ["<"], values = [3])
        self.__should_pass(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 1))
        self.__should_fail(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 5))
        
        test_filter = FilterChain(fields = [FilterChain.datatype + '.attribute_1'], 
                                  operations = ["in"], values = ['test_val'])
        self.__should_pass(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 'test'))
        self.__should_fail(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 'test_bla'))
        
        test_filter = FilterChain(fields = [FilterChain.datatype + '.attribute_1'], 
                                  operations = ["in"], values = [['test_val', 'other_val']])
        self.__should_pass(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 'test_val'))
        self.__should_fail(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 'test_bla'))
        
        test_filter = FilterChain(fields = [FilterChain.datatype + '.attribute_1'], 
                                  operations = ["not in"], values = ['test_val'])
        self.__should_pass(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 'valoare'))
        self.__should_fail(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 'val'))
        
        test_filter = FilterChain(fields = [FilterChain.datatype + '.attribute_1'], 
                                  operations = ["not in"], values = [['test_val', 'other']])
        self.__should_pass(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 'taest_val'))
        self.__should_fail(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 'test_val'))


    def test_invalid_filter(self):
        """
        Error test-case when evaluating filter in Python.
        """
        test_filter = FilterChain(fields = [FilterChain.datatype + '.attribute_1'], 
                                  operations = ["in"], values = [None])
        self.assertRaises(InvalidFilterEntity, test_filter.get_python_filter_equivalent, 
                          FilteringTest.DummyFilterClass(attribute_1 = ['test_val', 'test2']))

        
    def test_invalid_input(self):
        """
        Error test-case.
        """
        test_filter = FilterChain(fields = [FilterChain.datatype + '.other_attribute_1'], 
                                  operations = ["in"], values = ['test'])
        self.assertRaises(InvalidFilterChainInput, test_filter.get_python_filter_equivalent, 
                          FilteringTest.DummyFilterClass(attribute_1 = ['test_val', 'test2']))

        
    def test_complex_filter(self):
        """
        Test a filter with at least 2 conditions
        """
        test_filter = FilterChain(fields= [FilterChain.datatype+ '.attribute_1', FilterChain.datatype+ '.attribute_2'],
                                  operations = ["==", 'in'], values = ['test_val', ['test_val2', 1]])
        self.__should_pass(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 'test_val', attribute_2 = 1))
        self.__should_pass(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 'test_val', attribute_2 = 1))
        self.__should_fail(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 'test_val', attribute_2 = 2))
        self.__should_fail(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 'test_val1', attribute_2 = 1))

    
    def test_filter_add_condition(self):
        """
        Test that adding a condition to a filter is working.
        """
        test_filter = FilterChain(fields = [FilterChain.datatype + '.attribute_1'], 
                                  operations = ["=="], values = ['test_val'])
        filter_input = FilteringTest.DummyFilterClass(attribute_1 = 'test_val', attribute_2 = 1)
        self.__should_pass(test_filter, filter_input)
        test_filter.add_condition(FilterChain.datatype + '.attribute_2', '==', 2)
        self.__should_fail(test_filter, filter_input)

    
    def test_filter_addition(self):
        """
        test addition in filter chain
        """
        filter1 = FilterChain(fields = [FilterChain.datatype + '.attribute_1'], 
                              operations = ["=="], values = ['test_val'])
        filter2 = FilterChain(fields = [FilterChain.datatype + '.attribute_2'], 
                              operations = ['in'], values = [['test_val2', 1]])
        test_filter = filter1 + filter2
        self.__should_pass(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 'test_val', attribute_2 = 1))
        self.__should_pass(test_filter, FilteringTest.DummyFilterClass(attribute_1='test_val', attribute_2='test_val2'))
        self.__should_fail(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 'test_val', attribute_2 = 2))
        self.__should_fail(test_filter, FilteringTest.DummyFilterClass(attribute_1 = 'test_val1', attribute_2 = 1))

    
def suite():
    """
    Gather all the tests in a test suite.
    """
    test_suite = unittest.TestSuite()
    test_suite.addTest(unittest.makeSuite(FilteringTest))
    return test_suite


if __name__ == "__main__":
    #So you can run tests from this package individually.
    TEST_RUNNER = unittest.TextTestRunner()
    TEST_SUITE = suite()
    TEST_RUNNER.run(TEST_SUITE)
    
    


# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
The TimeSeries datatypes. This brings together the scientific and framework 
methods that are associated with the time-series data.

.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>

"""

import json
import numpy
from tvb.basic.traits import core, types_basic as basic, exceptions, types_mapped
from tvb.datatypes import sensors, surfaces, volumes, region_mapping, connectivity, arrays
from tvb.basic.arguments_serialisation import (preprocess_space_parameters, preprocess_time_parameters,
    postprocess_voxel_ts)
from tvb.basic.logger.builder import get_logger

LOG = get_logger(__name__)


def prepare_time_slice(total_time_length, max_length=10 ** 4):
    """
    Limit the time dimension when retrieving from TS.
    If total time length is greater than MAX, then retrieve only the last part of the TS

    :param total_time_length: TS time dimension
    :param max_length: limiting number of TS steps

    :return: python slice
    """

    if total_time_length < max_length:
        return slice(total_time_length)

    return slice(total_time_length - max_length, total_time_length)


class TimeSeries(types_mapped.MappedType):
    """
    Base time-series dataType.
    """

    title = basic.String

    data = arrays.FloatArray(
        label="Time-series data",
        file_storage=core.FILE_STORAGE_EXPAND,
        doc="""An array of time-series data, with a shape of [tpts, :], where ':' represents 1 or more dimensions""")

    nr_dimensions = basic.Integer(
        label="Number of dimension in timeseries",
        default=4)

    length_1d, length_2d, length_3d, length_4d = [basic.Integer] * 4

    labels_ordering = basic.List(
        default=["Time", "State Variable", "Space", "Mode"],
        label="Dimension Names",
        doc="""List of strings representing names of each data dimension""")

    labels_dimensions = basic.Dict(
        default={},
        label="Specific labels for each dimension for the data stored in this timeseries.",
        doc=""" A dictionary containing mappings of the form {'dimension_name' : [labels for this dimension] }""")

    time = arrays.FloatArray(
        file_storage=core.FILE_STORAGE_EXPAND,
        label="Time-series time",
        required=False,
        doc="""An array of time values for the time-series, with a shape of [tpts,].
        This is 'time' as returned by the simulator's monitors.""")

    start_time = basic.Float(label="Start Time:")

    sample_period = basic.Float(label="Sample period", default=1.0)

    # Specify the measure unit for sample period (e.g sec, msec, usec, ...)
    sample_period_unit = basic.String(
        label="Sample Period Measure Unit",
        default="ms")

    sample_rate = basic.Float(
        label="Sample rate",
        doc="""The sample rate of the timeseries""")

    has_surface_mapping = basic.Bool(default=True)
    has_volume_mapping = basic.Bool(default=False)

    def configure(self):
        """
        After populating few fields, compute the rest of the fields
        """
        super(TimeSeries, self).configure()
        data_shape = self.read_data_shape()
        self.nr_dimensions = len(data_shape)
        self.sample_rate = 1.0 / self.sample_period

        for i in range(min(self.nr_dimensions, 4)):
            setattr(self, 'length_%dd' % (i + 1), int(data_shape[i]))

    def read_data_shape(self):
        """
        Expose shape read on field data.
        """
        try:
            return self.get_data_shape('data')
        except exceptions.TVBException:
            self.logger.exception("Could not read data shape for TS!")
            raise exceptions.TVBException("Invalid empty TimeSeries!")

    def read_data_slice(self, data_slice):
        """
        Expose chunked-data access.
        """
        return self.get_data('data', data_slice)

    def read_time_page(self, current_page, page_size, max_size=None):
        """
        Compute time for current page.
        :param current_page: Starting from 0
        """
        current_page = int(current_page)
        page_size = int(page_size)

        if max_size is None:
            max_size = page_size
        else:
            max_size = int(max_size)

        page_real_size = page_size * self.sample_period
        start_time = self.start_time + current_page * page_real_size
        end_time = start_time + min(page_real_size, max_size * self.sample_period)

        return numpy.arange(start_time, end_time, self.sample_period)

    def read_channels_page(self, from_idx, to_idx, step=None, specific_slices=None, channels_list=None):
        """
        Read and return only the data page for the specified channels list.

        :param from_idx: the starting time idx from which to read data
        :param to_idx: the end time idx up until to which you read data
        :param step: increments in which to read the data. Optional, default to 1.
        :param specific_slices: optional parameter. If speficied slices the data accordingly.
        :param channels_list: the list of channels for which we want data
        """
        if channels_list:
            channels_list = json.loads(channels_list)
            for i in range(len(channels_list)):
                channels_list[i] = int(channels_list[i])

        if channels_list:
            channel_slice = tuple(channels_list)
        else:
            channel_slice = slice(None)

        data_page = self.read_data_page(from_idx, to_idx, step, specific_slices)
        # This is just a 1D array like in the case of Global Average monitor.
        # No need for the channels list
        if len(data_page.shape) == 1:
            return data_page.reshape(data_page.shape[0], 1)
        else:
            return data_page[:, channel_slice]

    def read_data_page(self, from_idx, to_idx, step=None, specific_slices=None):
        """
        Retrieve one page of data (paging done based on time).
        """
        from_idx, to_idx = int(from_idx), int(to_idx)

        if isinstance(specific_slices, basestring):
            specific_slices = json.loads(specific_slices)
        if step is None:
            step = 1
        else:
            step = int(step)

        slices = []
        overall_shape = self.read_data_shape()
        for i in range(len(overall_shape)):
            if i == 0:
                # Time slice
                slices.append(
                    slice(from_idx, min(to_idx, overall_shape[0]), step))
                continue
            if i == 2:
                # Read full of the main_dimension (space for the simulator)
                slices.append(slice(overall_shape[i]))
                continue
            if specific_slices is None:
                slices.append(slice(0, 1))
            else:
                slices.append(slice(specific_slices[i], min(specific_slices[i] + 1, overall_shape[i]), 1))

        data = self.read_data_slice(tuple(slices))
        if len(data) == 1:
            # Do not allow time dimension to get squeezed, a 2D result need to
            # come out of this method.
            data = data.squeeze()
            data = data.reshape((1, len(data)))
        else:
            data = data.squeeze()

        return data

    def read_data_page_split(self, from_idx, to_idx, step=None, specific_slices=None):
        """
        No Split needed in case of basic TS (sensors and region level)
        """
        return self.read_data_page(from_idx, to_idx, step, specific_slices)


    def write_time_slice(self, partial_result):
        """
        Append a new value to the ``time`` attribute.
        """
        self.store_data_chunk("time", partial_result, grow_dimension=0, close_file=False)

    def write_data_slice(self, partial_result, grow_dimension=0):
        """
        Append a chunk of time-series data to the ``data`` attribute.
        """
        self.store_data_chunk("data", partial_result, grow_dimension=grow_dimension, close_file=False)

    def get_min_max_values(self):
        """
        Retrieve the minimum and maximum values from the metadata.
        :returns: (minimum_value, maximum_value)
        """
        metadata = self.get_metadata('data')
        return metadata[self.METADATA_ARRAY_MIN], metadata[self.METADATA_ARRAY_MAX]

    def get_space_labels(self):
        """
        It assumes that we want to select in the 3'rd dimension,
        and generates labels for each point in that dimension.
        Subclasses are more specific.
        :return: An array of strings.
        """
        if self.nr_dimensions > 2:
            return ['signal-%d' % i for i in range(self._length_3d)]
        else:
            return []

    def get_grouped_space_labels(self):
        """
        :return: A list of label groups. A label group is a tuple (name, [(label_idx, label)...]).
                 Default all labels in a group named ''
        """
        return [('', list(enumerate(self.get_space_labels())))]

    def get_default_selection(self):
        """
        :return: The measure point indices that have to be shown by default. By default show all.
        """
        return range(len(self.get_space_labels()))

    def get_measure_points_selection_gid(self):
        """
        :return: a datatype gid with which to obtain al valid measure point selection for this time series
                 We have to decide if the default should be all selections or none
        """
        return ''

    @staticmethod
    def accepted_filters():
        filters = types_mapped.MappedType.accepted_filters()
        filters.update({'datatype_class._nr_dimensions': {'type': 'int', 'display': 'No of Dimensions',
                                                          'operations': ['==', '<', '>']},
                        'datatype_class._sample_period': {'type': 'float', 'display': 'Sample Period',
                                                          'operations': ['==', '<', '>']},
                        'datatype_class._sample_rate': {'type': 'float', 'display': 'Sample Rate',
                                                        'operations': ['==', '<', '>']},
                        'datatype_class._title': {'type': 'string', 'display': 'Title',
                                                  'operations': ['==', '!=', 'like']}})
        return filters

    def _find_summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        summary = {"Time-series type": self.__class__.__name__,
                   "Time-series name": self.title,
                   "Dimensions": self.labels_ordering,
                   "Time units": self.sample_period_unit,
                   "Sample period": self.sample_period,
                   "Length": self.sample_period * self.get_data_shape('data')[0]}
        summary.update(self.get_info_about_array('data'))
        return summary


class SensorsTSBase(TimeSeries):
    """
    Add framework related functionality for TS Sensor classes

    """
    __tablename__ = None


    def get_space_labels(self):
        """
        :return: An array of strings with the sensors labels.
        """
        if self.sensors is not None:
            return list(self.sensors.labels)
        return []

    def get_measure_points_selection_gid(self):
        if self.sensors is not None:
            return self.sensors.gid
        return ''

    def get_default_selection(self):
        if self.sensors is not None:
            # select only the first 8 channels
            return range(min(8, len(self.get_space_labels())))
        return []

    def _find_summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        summary = super(SensorsTSBase, self)._find_summary_info()
        summary.update({"Source Sensors": self.sensors.display_name})
        return summary


class TimeSeriesEEG(SensorsTSBase):
    """ A time series associated with a set of EEG sensors. """
    _ui_name = "EEG time-series"
    __generate_table__ = True

    sensors = sensors.SensorsEEG
    labels_ordering = basic.List(default=["Time", "1", "EEG Sensor", "1"])


class TimeSeriesMEG(SensorsTSBase):
    """ A time series associated with a set of MEG sensors. """
    _ui_name = "MEG time-series"
    __generate_table__ = True

    sensors = sensors.SensorsMEG
    labels_ordering = basic.List(default=["Time", "1", "MEG Sensor", "1"])


class TimeSeriesSEEG(SensorsTSBase):
    """ A time series associated with a set of Internal sensors. """
    _ui_name = "Stereo-EEG time-series"
    __generate_table__ = True

    sensors = sensors.SensorsInternal
    labels_ordering = basic.List(default=["Time", "1", "sEEG Sensor", "1"])


class TimeSeriesRegion(TimeSeries):
    """ A time-series associated with the regions of a connectivity. """
    _ui_name = "Region time-series"
    connectivity = connectivity.Connectivity
    region_mapping_volume = region_mapping.RegionVolumeMapping(required=False)
    region_mapping = region_mapping.RegionMapping(required=False)
    labels_ordering = basic.List(default=["Time", "State Variable", "Region", "Mode"])

    def configure(self):
        """
        After populating few fields, compute the rest of the fields
        """
        super(TimeSeriesRegion, self).configure()
        self.has_surface_mapping = self.region_mapping is not None or self._region_mapping is not None
        self.has_volume_mapping = self.region_mapping_volume is not None or self._region_mapping_volume is not None

    def _find_summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        summary = super(TimeSeriesRegion, self)._find_summary_info()
        summary.update({"Source Connectivity": self.connectivity.display_name,
                        "Region Mapping": self.region_mapping.display_name if self.region_mapping else "None",
                        "Region Mapping Volume": (self.region_mapping_volume.display_name
                                                  if self.region_mapping_volume else "None")})
        return summary

    def get_space_labels(self):
        """
        :return: An array of strings with the connectivity node labels.
        """
        if self.connectivity is not None:
            return list(self.connectivity.region_labels)
        return []

    def get_grouped_space_labels(self):
        """
        :return: A structure of this form [('left', [(idx, lh_label)...]), ('right': [(idx, rh_label) ...])]
        """
        if self.connectivity is not None:
            return self.connectivity.get_grouped_space_labels()
        else:
            return super(TimeSeriesRegion, self).get_grouped_space_labels()

    def get_default_selection(self):
        """
        :return: If the connectivity of this time series is edited from another
                 return the nodes of the parent that are present in the connectivity.
        """
        if self.connectivity is not None:
            return self.connectivity.get_default_selection()
        else:
            return super(TimeSeriesRegion, self).get_default_selection()

    def get_measure_points_selection_gid(self):
        """
        :return: the associated connectivity gid
        """
        if self.connectivity is not None:
            return self.connectivity.get_measure_points_selection_gid()
        else:
            return super(TimeSeriesRegion, self).get_measure_points_selection_gid()

    def get_volume_view(self, from_idx, to_idx, x_plane, y_plane, z_plane, var=0, mode=0):
        """
        Retrieve 3 slices through the Volume TS, at the given X, y and Z coordinates, and in time [from_idx .. to_idx].

        :param from_idx: int This will be the limit on the first dimension (time)
        :param to_idx: int Also limit on the first Dimension (time)
        :param x_plane: int coordinate
        :param y_plane: int coordinate
        :param z_plane: int coordinate

        :return: An array of 3 Matrices 2D, each containing the values to display in planes xy, yz and xy.
        """

        if self.region_mapping_volume is None:
            raise exceptions.TVBException("Invalid method called for TS without Volume Mapping!")

        volume_rm = self.region_mapping_volume

        # Work with space inside Volume:
        x_plane, y_plane, z_plane = preprocess_space_parameters(x_plane, y_plane, z_plane, volume_rm.length_1d,
                                                                volume_rm.length_2d, volume_rm.length_3d)

        slice_x, slice_y, slice_z = self.region_mapping_volume.get_volume_slice(x_plane, y_plane, z_plane)

        # Read from the current TS:
        from_idx, to_idx, current_time_length = preprocess_time_parameters(from_idx, to_idx, self.read_data_shape()[0])
        no_of_regions = self.read_data_shape()[2]
        time_slices = slice(from_idx, to_idx), slice(var, var + 1), slice(no_of_regions), slice(mode, mode + 1)

        min_signal = self.get_min_max_values()[0]
        regions_ts = self.read_data_slice(time_slices)[:, 0, :, 0]
        regions_ts = numpy.hstack((regions_ts, numpy.ones((current_time_length, 1)) * self.out_of_range(min_signal)))

        # Index from TS with the space mapping:
        result_x, result_y, result_z = [], [], []

        for i in range(0, current_time_length):
            result_x.append(regions_ts[i][slice_x].tolist())
            result_y.append(regions_ts[i][slice_y].tolist())
            result_z.append(regions_ts[i][slice_z].tolist())

        return [result_x, result_y, result_z]

    def get_voxel_time_series(self, x, y, z, var=0, mode=0):
        """
        Retrieve for a given voxel (x,y,z) the entire timeline.

        :param x: int coordinate
        :param y: int coordinate
        :param z: int coordinate

        :return: A complex dictionary with information about current voxel.
                The main part will be a vector with all the values over time from the x,y,z coordinates.
        """

        if self.region_mapping_volume is None:
            raise exceptions.TVBException("Invalid method called for TS without Volume Mapping!")

        volume_rm = self.region_mapping_volume
        x, y, z = preprocess_space_parameters(x, y, z, volume_rm.length_1d, volume_rm.length_2d, volume_rm.length_3d)
        idx_slices = slice(x, x + 1), slice(y, y + 1), slice(z, z + 1)

        idx = int(volume_rm.get_data('array_data', idx_slices))

        time_length = self.read_data_shape()[0]
        voxel_slices = prepare_time_slice(time_length), slice(var, var + 1), slice(idx, idx + 1), slice(mode, mode + 1)
        label = volume_rm.connectivity.region_labels[idx]

        background, back_min, back_max = None, None, None
        if idx < 0:
            back_min, back_max = self.get_min_max_values()
            background = numpy.ones((time_length, 1)) * self.out_of_range(back_min)
            label = 'background'

        result = postprocess_voxel_ts(self, voxel_slices, background, back_min, back_max, label)
        return result

    @staticmethod
    def out_of_range(min_value):
        return round(min_value) - 1


class TimeSeriesSurface(TimeSeries):
    """ A time-series associated with a Surface. """
    _ui_name = "Surface time-series"
    surface = surfaces.CorticalSurface
    labels_ordering = basic.List(default=["Time", "State Variable", "Vertex", "Mode"])
    SELECTION_LIMIT = 100

    def get_space_labels(self):
        """
        Return only the first `SELECTION_LIMIT` vertices/channels
        """
        return ['signal-%d' % i for i in range(min(self._length_3d, self.SELECTION_LIMIT))]

    def _find_summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        summary = super(TimeSeriesSurface, self)._find_summary_info()
        summary.update({"Source Surface": self.surface.display_name})
        return summary

    def read_data_page_split(self, from_idx, to_idx, step=None, specific_slices=None):

        basic_result = self.read_data_page(from_idx, to_idx, step, specific_slices)
        result = []
        if self.surface.number_of_split_slices <= 1:
            result.append(basic_result)
        else:
            for slice_number in range(self.surface.number_of_split_slices):
                start_idx, end_idx = self.surface._get_slice_vertex_boundaries(slice_number)
                result.append(basic_result[:,start_idx:end_idx].tolist())

        return result


class TimeSeriesVolume(TimeSeries):
    """ A time-series associated with a Volume. """
    _ui_name = "Volume time-series"
    volume = volumes.Volume
    labels_ordering = basic.List(default=["Time", "X", "Y", "Z"])

    def _find_summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        summary = super(TimeSeriesVolume, self)._find_summary_info()
        summary.update({"Source Volume": self.volume.display_name})
        return summary

    def configure(self):
        """
        After populating few fields, compute the rest of the fields
        """
        super(TimeSeriesVolume, self).configure()
        self.has_volume_mapping = True

    def get_volume_view(self, from_idx, to_idx, x_plane, y_plane, z_plane):
        """
        Retrieve 3 slices through the Volume TS, at the given X, y and Z coordinates, and in time [from_idx .. to_idx].

        :param from_idx: int This will be the limit on the first dimension (time)
        :param to_idx: int Also limit on the first Dimension (time)
        :param x_plane: int coordinate
        :param y_plane: int coordinate
        :param z_plane: int coordinate

        :return: An array of 3 Matrices 2D, each containing the values to display in planes xy, yz and xy.
        """

        overall_shape = self.read_data_shape()
        from_idx, to_idx, time = preprocess_time_parameters(from_idx, to_idx, overall_shape[0])
        x_plane, y_plane, z_plane = preprocess_space_parameters(x_plane, y_plane, z_plane,
                                                                overall_shape[1], overall_shape[2], overall_shape[3])

        slices = slice(from_idx, to_idx), slice(overall_shape[1]), slice(overall_shape[2]), slice(z_plane, z_plane + 1)
        slicex = self.read_data_slice(slices)[:, :, :, 0].tolist()

        slices = slice(from_idx, to_idx), slice(x_plane, x_plane + 1), slice(overall_shape[2]), slice(overall_shape[3])
        slicey = self.read_data_slice(slices)[:, 0, :, :][..., ::-1].tolist()

        slices = slice(from_idx, to_idx), slice(overall_shape[1]), slice(y_plane, y_plane + 1), slice(overall_shape[3])
        slicez = self.read_data_slice(slices)[:, :, 0, :][..., ::-1].tolist()

        return [slicex, slicey, slicez]

    def get_voxel_time_series(self, x, y, z):
        """
        Retrieve for a given voxel (x,y,z) the entire timeline.

        :param x: int coordinate
        :param y: int coordinate
        :param z: int coordinate

        :return: A complex dictionary with information about current voxel.
                The main part will be a vector with all the values over time from the x,y,z coordinates.
        """

        overall_shape = self.read_data_shape()
        x, y, z = preprocess_space_parameters(x, y, z, overall_shape[1], overall_shape[2], overall_shape[3])

        slices = prepare_time_slice(overall_shape[0]), slice(x, x + 1), slice(y, y + 1), slice(z, z + 1)

        result = postprocess_voxel_ts(self, slices)
        return result



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""  
Calculate temporal cross correlation coefficients (Pearson's
coefficient) on a TimeSeries datatype and return a
graph.CorrelationCoefficients dataype. The correlation matrix is
widely used to represent functional connectivity (FC).

.. moduleauthor:: Paula Sanz Leon <paula.sanz-leon@univ-amu.fr>

"""

import numpy
#TODO: Currently built around the Simulator's 4D timeseries -- generalise...
import tvb.datatypes.time_series as time_series
import tvb.datatypes.graph as graph
import tvb.basic.traits.core as core
import tvb.basic.traits.types_basic as basic
import tvb.basic.traits.util as util
from tvb.basic.logger.builder import get_logger

LOG = get_logger(__name__)




class CorrelationCoefficient(core.Type):
    """
    Compute the node-pairwise pearson correlation coefficient of the
    given input 4D TimeSeries  datatype.
    
    Return a CrossCorrelation datatype, whose values of are between -1
    and 1, inclusive.
    
    See: http://docs.scipy.org/doc/numpy/reference/generated/numpy.corrcoef.html
    """

    time_series = time_series.TimeSeries(
        label = "Time Series",
        required = True,
        doc = """The time-series for which the cross correlation matrices are
        calculated.""")

    t_start = basic.Float(
        label = ":math:`t_{start}`",
        default = 0.9765625,
        required = True,
        doc = """Time start point (ms). By default it uses the default Monitor sample period.
        The starting time point of a time series is not zero, but the monitor's sample period. """)

    t_end = basic.Float(
        label = ":math:`t_{end}`",
        default = 1000.,
        required = True,
        doc = """ End time point (ms) """)


    def evaluate(self):
        """
        Compute the correlation coefficients of a 2D array (tpts x nodes).
        Yields an array of size nodes x nodes x state-variables x modes.

        The time interval over which the correlation coefficients are computed 
        is defined by t_start, t_end

        """
        cls_attr_name = self.__class__.__name__ + ".time_series"
        self.time_series.trait["data"].log_debug(owner=cls_attr_name)

        #(nodes, nodes, state-variables, modes)
        input_shape = self.time_series.read_data_shape()
        result_shape = self.result_shape(input_shape)
        LOG.info("result shape will be: %s" % str(result_shape))

        result = numpy.zeros(result_shape)


        t_lo = int((1. / self.time_series.sample_period) * (self.t_start - self.time_series.sample_period))
        t_hi = int((1. / self.time_series.sample_period) * (self.t_end - self.time_series.sample_period))
        t_lo = max(t_lo, 0)
        t_hi = max(t_hi, input_shape[0])

        #One correlation coeff matrix, for each state-var & mode.
        for mode in range(result_shape[3]):
            for var in range(result_shape[2]):
                current_slice = tuple([slice(t_lo, t_hi + 1), slice(var, var + 1),
                                       slice(input_shape[2]), slice(mode, mode + 1)])
                data = self.time_series.read_data_slice(current_slice).squeeze()
                result[:, :, var, mode] = numpy.corrcoef(data.T)


        util.log_debug_array(LOG, result, "result")

        corr_coeff = graph.CorrelationCoefficients(source=self.time_series,
                                                   array_data=result,
                                                   use_storage=False)
        return corr_coeff


    def result_shape(self, input_shape):
        """Returns the shape of the main result of ...."""
        result_shape = (input_shape[2], input_shape[2],
                        input_shape[1], input_shape[3])
        return result_shape


    def result_size(self, input_shape):
        """
        Returns the storage size in Bytes of the main result of .
        """
        result_size = numpy.sum(map(numpy.prod, self.result_shape(input_shape))) * 8.0  # Bytes
        return result_size


    def extended_result_size(self, input_shape):
        """
        Returns the storage size in Bytes of the extended result of the ....
        That is, it includes storage of the evaluated ... attributes
        such as ..., etc.
        """
        extend_size = self.result_size(input_shape)  # Currently no derived attributes.
        return extend_size






# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""

The RegionBoundaries class, given a cortical surface with region_mapping based
on a particular parcellation, provides access to region neighbours and 
boundaries within the surface.

.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>

"""
#TODO: debug, clean, doc
#
import numpy

#The Virtual Brain

from tvb.simulator.common import get_logger
LOG = get_logger(__name__)

#import tvb.basic.traits as trait
#import tvb.simulator.surfaces as surfaces_module


class RegionBoundaries(object):
    """
    """
#TODO: Cortex attribute isn't properly initialising -- doing something wrong...    
#    cortex = trait.Instance(label = "A cortical surface object",
#                            default = surfaces_module.Cortex(),
#                            default_class = surfaces_module.Cortex,
#                            value = surfaces_module.Cortex(),  #trait_type
#                            datatype = True,
#                            required = True)

    def __init__(self, cortex): # **kwargs
        """
        """

#        super(RegionBoundaries, self).__init__(**kwargs)
        self.cortex = cortex
        self._boundary_triangles = None # All triangles that cross a boundary
        self._boundary_edges = None # All edges that cross a boundary
        self._neighbouring_regions = None
        self._region_neighbours = None # Dict of region neihbours
        self._region_adjacency= None # Adjacency matrix indicating region neighbours
        self._region_boundaries = None # Vertices at the edge of each region
        self._boundary = None


    @property
    def boundary_triangles(self):
        """ The ... """
        if self._boundary_triangles is None:
           self._boundary_triangles = find_boundary_triangles(self.cortex)
        return self._boundary_triangles


    @property
    def boundary_edges(self):
        """ The ... """
        if self._boundary_edges is None:
           self._boundary_edges = find_boundary_edges(self.cortex)
        return self._boundary_edges


    @property
    def neighbouring_regions(self):
        """ The ... """
        if self._neighbouring_regions is None:
           self._neighbouring_regions = find_neighbouring_regions(self.cortex, self.boundary_edges)
        return self._neighbouring_regions


    @property
    def region_neighbours(self):
        """ The ... """
        if self._region_neighbours is None:
           self._region_neighbours = find_region_neighbours(self.region_adjacency)
        return self._region_neighbours


    @property
    def number_of_neighbours(self):
        """ The ... """
        if self._number_of_neighbours is None:
           self._number_of_neighbours = find_number_of_neighbours(self.region_neighbours)
        return self._number_of_neighbours


    @property
    def region_adjacency(self):
        """ The ... """
        if self._region_adjacency is None:
           self._region_adjacency = find_region_adjacency(self.neighbouring_regions)
        return self._region_adjacency


    @property
    def region_boundaries(self):
        """ The ... """
        if self._region_boundaries is None:
           self._region_boundaries = find_region_boundaries(self.cortex)
        return self._region_boundaries


    @property
    def boundary(self):
        """ The ... """
        if self._boundary is None:
           self._boundary = find_boundary(self.cortex, self.boundary_edges)
        return self._boundary



def find_boundary_triangles(cortex):
    """ 
    Identify triangles that cross a parcellation boundary 
    """
    tb01 = numpy.nonzero(cortex.region_mapping[cortex.triangles][:, 0] - 
                         cortex.region_mapping[cortex.triangles][:, 1])

    tb12 = numpy.nonzero(cortex.region_mapping[cortex.triangles][:, 1] - 
                         cortex.region_mapping[cortex.triangles][:, 2])

    tb20 = numpy.nonzero(cortex.region_mapping[cortex.triangles][:, 2] - 
                         cortex.region_mapping[cortex.triangles][:, 0])

    return numpy.unique(numpy.hstack((tb01, tb12, tb20)))


def find_boundary_edges(cortex):
    """ 
    Identify edges that cross a parcellation boundary 
    """
    tb01 = numpy.nonzero(cortex.region_mapping[cortex.triangles][:, 0] -
                         cortex.region_mapping[cortex.triangles][:, 1])

    tb12 = numpy.nonzero(cortex.region_mapping[cortex.triangles][:, 1] -
                         cortex.region_mapping[cortex.triangles][:, 2])

    tb20 = numpy.nonzero(cortex.region_mapping[cortex.triangles][:, 2] -
                         cortex.region_mapping[cortex.triangles][:, 0])
    ed01 = numpy.vstack((cortex.triangles[tb01, 0], cortex.triangles[tb01, 1]))
    ed12 = numpy.vstack((cortex.triangles[tb12, 1], cortex.triangles[tb12, 2]))
    ed20 = numpy.vstack((cortex.triangles[tb20, 2], cortex.triangles[tb20, 0]))

    all_boundary_edges = numpy.hstack((ed01, ed12, ed20)).T
    all_boundary_edges.sort()
    boundary_edge_set = set(map(tuple, all_boundary_edges.tolist()))
    unique_boundary_edges = numpy.zeros((2, len(boundary_edge_set)), dtype=numpy.int32)
    k =0
    while boundary_edge_set:
        unique_boundary_edges[:, k] = numpy.array(boundary_edge_set.pop())
        k+=1
    return unique_boundary_edges


def find_region_boundaries(cortex, boundary_triangles):
    """ """
    boundary_vertices = numpy.unique(cortex.triangles[boundary_triangles, :])
    vertex_regions = cortex.region_mapping[boundary_vertices]
    region_boundaries = []
    for k in numpy.unique(cortex.region_mapping):
        region_boundaries.append(boundary_vertices[vertex_regions==k])
    return region_boundaries


def find_neighbouring_regions(cortex, boundary_edges):
    """ """
    nr = cortex.region_mapping[boundary_edges]
    nr = nr.T
    nr.sort()
    neighbouring_region_set = set(map(tuple, nr.tolist()))
    unique_neighbouring_regions = numpy.zeros((2, len(neighbouring_region_set)), dtype=numpy.int32)
    k = 0
    while neighbouring_region_set:
        unique_neighbouring_regions[:, k] = numpy.array(neighbouring_region_set.pop())
        k+=1
    return unique_neighbouring_regions


def find_boundary(cortex, boundary_edges):
    """
    """
    edge_vert_pairs = cortex.vertices[boundary_edges, :]
    vec_0to1 = (edge_vert_pairs[1, :] - edge_vert_pairs[0, :])
    boundary = edge_vert_pairs[0, :] +  vec_0to1 / 2.0
    return boundary


def find_region_adjacency(neighbouring_regions):
    """
    """
    number_of_regions = neighbouring_regions.max() + 1
    region_adjacency = numpy.zeros((number_of_regions, number_of_regions), 
                                   dtype=numpy.int32)

    region_adjacency[neighbouring_regions[0], neighbouring_regions[1]] = 1

    return region_adjacency



def find_region_neighbours(region_adjacency):
    """
    """
    #NOTE: 'should only be doing 1 hemisphere here and then flipping, for symmetry.
    number_of_regions = region_adjacency.shape[0]
    xxx = numpy.nonzero(region_adjacency + region_adjacency.T)
    neighbours = {}
    for key in range(number_of_regions):
        #Assign 
        neighbours[key] = list(xxx[1][xxx[0]==key])
#        # Extend neighbours to include "colourbar neighbours"
#        neighbours[key].append(numpy.mod(key+1, number_of_regions))
#        neighbours[key].append(numpy.mod(key-1, number_of_regions))
    return neighbours



def find_number_of_neighbours(neighbours):
    """
    """
    number_of_regions = len(neighbours)
    number_of_neighbours = numpy.zeros(number_of_regions, dtype=numpy.uint8)
    for k in range(number_of_regions):
        number_of_neighbours[k] = len(neighbours[k])


### EoF ###



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
"""
Created on Mar 20, 2013

.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
"""
if __name__ == "__main__":
    from tvb.tests.library import setup_test_console_env
    setup_test_console_env()
 
import numpy   
import unittest

from tvb.datatypes import mode_decompositions, time_series
from tvb.tests.library.base_testcase import BaseTestCase
        
class ModeDecompositionsTest(BaseTestCase):
    """
    Tests the defaults for `tvb.datatypes.mode_decompositions` module.
    """
    
    def test_principalcomponents(self):
        data = numpy.random.random((10, 10, 10, 10))
        ts = time_series.TimeSeries(data=data)
        dt = mode_decompositions.PrincipalComponents(source = ts,
                                                    fractions = numpy.random.random((10, 10, 10)),
                                                    weights = data)
        dt.configure()
        dt.compute_norm_source()
        dt.compute_component_time_series()
        dt.compute_normalised_component_time_series()
        summary = dt.summary_info
        self.assertEqual(summary['Mode decomposition type'], 'PrincipalComponents')
        self.assertTrue(dt.source is not None)
        self.assertEqual(dt.weights.shape, (10, 10, 10, 10))
        self.assertEqual(dt.fractions.shape, (10, 10, 10))
        self.assertEqual(dt.norm_source.shape, (10, 10, 10, 10))
        self.assertEqual(dt.component_time_series.shape, (10, 10, 10, 10))
        self.assertEqual(dt.normalised_component_time_series.shape, (10, 10, 10, 10))
        
        
    def test_independentcomponents(self):
        data = numpy.random.random((10, 10, 10, 10))
        ts = time_series.TimeSeries(data=data)
        n_comp = 5
        dt = mode_decompositions.IndependentComponents(  source = ts,
                                         component_time_series = numpy.random.random((10, n_comp, 10, 10)), 
                                         prewhitening_matrix = numpy.random.random((n_comp, 10, 10, 10)),
                                         unmixing_matrix = numpy.random.random((n_comp, n_comp, 10, 10)),
                                         n_components = n_comp)
        dt.compute_norm_source()
        dt.compute_component_time_series()
        dt.compute_normalised_component_time_series()
        summary = dt.summary_info
        self.assertEqual(summary['Mode decomposition type'], 'IndependentComponents')
        self.assertTrue(dt.source is not None)
        self.assertEqual(dt.mixing_matrix.shape, (0,))
        self.assertEqual(dt.unmixing_matrix.shape, (n_comp, n_comp, 10, 10))
        self.assertEqual(dt.prewhitening_matrix.shape, (n_comp, 10, 10, 10))
        self.assertEqual(dt.norm_source.shape, (10, 10, 10, 10))
        self.assertEqual(dt.component_time_series.shape, (10, 10, n_comp, 10))
        
        
def suite():
    """
    Gather all the tests in a test suite.
    """
    test_suite = unittest.TestSuite()
    test_suite.addTest(unittest.makeSuite(ModeDecompositionsTest))
    return test_suite


if __name__ == "__main__":
    #So you can run tests from this package individually.
    TEST_RUNNER = unittest.TextTestRunner()
    TEST_SUITE = suite()
    TEST_RUNNER.run(TEST_SUITE) 


# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Generate Dictionary required by the Framework to generate UI from it.
Returned dictionary will be generated from  traited definition of attributes.


.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Stuart Knock <stuart.knock@gmail.com>
.. moduleauthor:: marmaduke <duke@eml.cc>

"""

import numpy
import json
from tvb.basic.logger.builder import get_logger
from tvb.basic.traits.util import get, str_class_name, multiline_math_directives_to_matjax
from tvb.basic.traits.core import KWARG_AVOID_SUBCLASSES, TYPE_REGISTER, KWARG_FILTERS_UI, KWARG_FILTERS_BACKEND
import itree_model as itr
LOG = get_logger(__name__)

INTERFACE_ATTRIBUTES_ONLY = "attributes-only"
INTERFACE_ATTRIBUTES = "attributes"



class TraitedInterfaceGeneratorExperimental(object):
    def __get__(self, inst, ownr):

        obj = inst if inst else ownr
        if not obj.trait.bound:
            return None

        name, label, description, default, required, locked = self.__get_basics(obj)

        mro_type_names = [i.__name__ for i in ownr.mro()]

        if 'Integer' in mro_type_names or 'Float' in mro_type_names:
            type_ = self.__get_wrapped_type(obj)
            range_ = self.__get_range(obj)
            intr = itr.NumericNode(name, type_, default, range_, label, description, required)
        elif 'Array' in mro_type_names:
            range_ = self.__get_range(obj)
            if isinstance(obj.trait.value, numpy.ndarray):
                # Make sure arrays are displayed in a compatible form: [1, 2, 3]
                default = str(obj.trait.value.tolist())
            intr = itr.ArrayNode(name, str(inst.dtype), default, range_, label, description, required)
        elif 'MappedType' in mro_type_names:
            if not ownr._ui_complex_datatype:
                ownr_class = ownr.__class__
            else:
                ownr_class = ownr._ui_complex_datatype
            if 'MetaType' in ownr_class.__name__:
                ownr_class = ownr().__class__
            type_ = ownr_class.__module__ + '.' + ownr_class.__name__

            if not ownr._ui_complex_datatype:
                intr = itr.DatatypeNode(name, type_, label, description, required)
            else:
                attributes = self.__get_entity_attributes(obj)
                intr = itr.ComplexDtypeNode(name, type_, attributes, label, description, required)

            if KWARG_FILTERS_UI in obj.trait.inits.kwd:
                intr.filters_ui = json.dumps([ui_filter.to_dict() for ui_filter in
                                                     obj.trait.inits.kwd[KWARG_FILTERS_UI]])
            if KWARG_FILTERS_BACKEND in obj.trait.inits.kwd:
                intr.conditions = obj.trait.inits.kwd[KWARG_FILTERS_BACKEND]
        elif 'Dict' in mro_type_names and isinstance(default, dict):
            attributes, elementType = self.__prepare_dictionary(default)
            if len(attributes) == 0:
                ## Dictionary without any sub-parameter
                return None
            intr = itr.DictNode(name, elementType, attributes, default)
        elif 'Enumerate' in mro_type_names:
            options = [itr.Option(val, val) for val in obj.trait.options]
            intr = itr.EnumerateNode(name, obj.trait.select_multiple, obj.trait.value, options,
                                     label, description, required)

        ##### TYPE & subclasses specifics ######################################
        elif ('Type' in mro_type_names and obj.__module__ != 'tvb.basic.traits.types_basic'
              or 'Range' in mro_type_names):

            if obj.trait.bound == INTERFACE_ATTRIBUTES_ONLY:
                # We need to do this, to avoid infinite loop on attributes
                # of class Type with no subclasses
                return self.__get_entity_attributes(obj)  # todo edge case if this is not called from __handle_nonmapped_subtypes

            intr = itr.SelectTypeNode(name, obj.trait.select_multiple, [], default, label, description)
            intr.options = self.__handle_nonmapped_subtypes(ownr, obj)

            self.__correct_defaults(intr)
        else:
            type_ = self.__get_wrapped_type(obj)
            intr = itr.LeafNode(name, type_)

        # nonstandard attributes
        self.__fill_noiseconfig(obj, intr)
        intr.locked = locked

        return intr


    @staticmethod
    def __get_basics(obj):
        label = get(obj.trait.inits.kwd, 'label', obj.trait.name)
        if not label:
            label = obj.trait.name
        default = (obj.value or obj.trait.value) if hasattr(obj, 'value') else obj.trait.value
        description = get(obj.trait.inits.kwd, 'doc')
        label = label.capitalize()
        name = obj.trait.name
        locked = obj.trait.inits.kwd.get('locked', False)
        required = obj.trait.inits.kwd.get('required', True)
        return name, label, description, default, required, locked


    @staticmethod
    def __get_range(obj):
        range_value = obj.trait.inits.kwd.get('range', False)
        if range_value:
            minValue = range_value.lo
            maxValue = range_value.hi
            if range_value.step is not None:
                stepValue = range_value.step
            else:
                LOG.debug("Missing Range.step field for attribute %s, we will consider a default." % obj.trait.name)
                stepValue = (range_value.hi - range_value.hi) / 10
            return itr.Range(minValue, maxValue, stepValue)


    @staticmethod
    def __fill_noiseconfig(obj, intr):
        noise_configurable = obj.trait.inits.kwd.get('configurable_noise', None)
        if noise_configurable is not None:
            intr.configurableNoise = noise_configurable


    @staticmethod
    def __get_wrapped_type(obj):
        if isinstance(obj.trait.wraps, tuple):
            return obj.trait.wraps[0].__name__
        else:
            return obj.trait.wraps.__name__


    @staticmethod
    def __get_entity_attributes(obj):
        # Populate Attributes for current entity
        attrs = sorted(obj.trait.values(), key=lambda entity: entity.trait.order_number)
        attrs = [val.interface_experimental for val in attrs if val.trait.order_number >= 0]
        return [attr for attr in attrs if attr is not None]


    def __prepare_dictionary(self, dictionary):
        """
        From base.Dict -> default [isinstance(dict)], prepare an interface specific tree.
        """
        result = []
        element_type = None
        for key in dictionary:
            value = dictionary[key]
            default = str(value)
            mro_type_names = [i.__name__ for i in type(value).mro()]
            if hasattr(value, 'tolist') or 'Array' in mro_type_names:
                type_ = 'array'
                if not hasattr(value, 'tolist'):
                    default = str(value.trait.value)
            else:
                type_ = type(value).__name__

            # assert type uniformity : element_type is type_
            element_type = type_
            if type_ == 'array':
                node = itr.ArrayNode(key, str(value.dtype), default, label=key)
            elif 'Integer' in mro_type_names or 'Float' in mro_type_names:
                node = itr.NumericNode(key, type_, default, label=key)
            else:
                node = itr.LeafNode(key, type_, default, key)
            result.append(node)
        return result, element_type


    @staticmethod
    def __handle_nonmapped_subtypes(ownr, obj):
        """ Populate options for each subtype. This fills in models etc"""
        options = []
        for opt in TYPE_REGISTER.subclasses(ownr, KWARG_AVOID_SUBCLASSES in obj.trait.inits.kwd):
            if hasattr(obj, 'value') and obj.value is not None and isinstance(obj.value, opt):
                ## fill option currently selected with attributes from instance
                opt = obj.value
                opt_class = opt.__class__
            else:
                opt_class = opt

            opt.trait.bound = INTERFACE_ATTRIBUTES_ONLY

            description = multiline_math_directives_to_matjax(opt_class.__doc__)
            name = get(opt, '_ui_name', opt_class.__name__)
            value=str_class_name(opt_class, short_form=True)
            type_=str_class_name(opt_class, short_form=False)
            attributes=opt.interface_experimental
            options.append(itr.TypeNode(name, value, type_, description, attributes))
        return options


    @staticmethod
    def __correct_defaults(intr):
        if intr.default is not None:
            intr.default = intr.default.__class__.__name__
            if intr.default == 'RandomState':
                intr.default = 'RandomStream'



    def __set__(self, inst, val):
        """
        Given a hierarchical dictionary of the kind generated by __get__, with the
        chosen options, we should be able to fully instantiate a class.
        """
        raise NotImplementedError




# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Test history in simulator.

.. moduleauthor:: Marmaduke Woodman <mw@eml.cc>

"""

if __name__ == "__main__":
    from tvb.tests.library import setup_test_console_env
    setup_test_console_env()

import numpy
import unittest
import tvb.basic.traits.types_basic as basic
from tvb.datatypes.connectivity import Connectivity
from tvb.simulator.coupling import Coupling
from tvb.simulator.integrators import Identity
from tvb.simulator.models import Model
from tvb.simulator.monitors import Raw
from tvb.simulator.simulator import Simulator
from tvb.tests.library.base_testcase import BaseTestCase



class IdCoupling(Coupling):
    """Implements an identity coupling function."""

    def __call__(self, step, history):
        g_ij = history.es_weights
        x_i, x_j = history.query(step)
        return (g_ij * x_j).sum(axis=2).transpose((1, 0, 2))



class Sum(Model):
    nvar = 1
    _nvar = 1
    state_variable_range = {'x': [0, 100]}
    variables_of_interest = basic.Enumerate(default=['x'], options=['x'])
    state_variables = ['x']
    cvar = numpy.array([0])

    def dfun(self, X, coupling, local_coupling=0):
        return X + coupling + local_coupling



class ExactPropagationTests(BaseTestCase):

    def build_simulator(self, n=4):

        self.conn = numpy.zeros((n, n), numpy.int32)
        for i in range(self.conn.shape[0] - 1):
            self.conn[i, i + 1] = 1

        self.dist = numpy.r_[:n * n].reshape((n, n))
        self.sim = Simulator(
            conduction_speed=1,
            coupling=IdCoupling(),
            surface=None,
            stimulus=None,
            integrator=Identity(dt=1),
            initial_conditions=numpy.ones((n * n, 1, n, 1)),
            simulation_length=10,
            connectivity=Connectivity(weights=self.conn, tract_lengths=self.dist, speed=1),
            model=Sum(),
            monitors=(Raw(), ),
        )
        self.sim.configure()


    def test_propagation(self):
        n = 4
        self.build_simulator(n=n)
        # x = numpy.zeros((n, ))
        xs = []
        for (t, raw), in self.sim(simulation_length=10):
            xs.append(raw.flat[:].copy())
        xs = numpy.array(xs)
        xs_ = numpy.array([[2., 2., 2., 1.],
                           [3., 3., 3., 1.],
                           [5., 4., 4., 1.],
                           [8., 5., 5., 1.],
                           [12., 6., 6., 1.],
                           [17., 7., 7., 1.],
                           [23., 8., 8., 1.],
                           [30., 10., 9., 1.],
                           [38., 13., 10., 1.],
                           [48., 17., 11., 1.]])
        self.assertTrue(numpy.allclose(xs, xs_))



def suite():
    """
    Gather all the tests in a test suite.
    """
    test_suite = unittest.TestSuite()
    test_suite.addTest(unittest.makeSuite(ExactPropagationTests))
    return test_suite



if __name__ == "__main__":
    #So you can run tests from this package individually.
    TEST_RUNNER = unittest.TextTestRunner()
    TEST_SUITE = suite()
    TEST_RUNNER.run(TEST_SUITE)



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
This module describes the simple of traited attributes one might needs on a class.

If your subclass should be mapped to a database table (true for most entities
that will be reused), use MappedType as superclass.

If you subclass is supported natively by SQLAlchemy, subclass Type, otherwise
subclass MappedType.

Important:
- Type - traited, possible mapped to db *col*
- MappedType - traited, mapped to db *table*


.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: marmaduke <duke@eml.cc>
.. moduleauthor:: Paula Sanz-Leon <paula.sanz-leon@univ-amu.fr>
"""

import json
import numpy
from decimal import Decimal
import tvb.basic.traits.core as core
from tvb.basic.logger.builder import get_logger

LOG = get_logger(__name__)


class String(core.Type):
    """
    Traits type that wraps a Python string.
    """
    wraps = (str, unicode)



class Bool(core.Type):
    """
    Traits type wrapping Python boolean primitive. 
    The only instances of Python bool are True and False.
    """
    wraps = bool


class Integer(core.Type):
    """
    Traits type that wraps Numpy's int32.
    """
    wraps = (int, long)



class Float(core.Type):
    """
    Traits type that wraps Numpy's float64.
    """
    wraps = (float, numpy.float32, int)


class Complex(core.Type):
    """
    Traits type that wraps Numpy's complex64.
    """
    wraps = numpy.complex64


class MapAsJson():
    """Add functionality of converting from/to JSON"""


    def __get__(self, inst, cls):
        if inst is not None and self.trait.bound and hasattr(inst, '_' + self.trait.name):

            if hasattr(inst, '__' + self.trait.name):
                return getattr(inst, '__' + self.trait.name)

            string = getattr(inst, '_' + self.trait.name)
            if string is None or (not isinstance(string, (str, unicode))):
                return string
            if len(string) < 1:
                return None
            json_value = self.from_json(string)

            # Cache for future usages (e.g. Stimulus.spatial should be the same instance on multiple accesses)
            setattr(inst, '__' + self.trait.name, json_value)
            return json_value
        return self


    @staticmethod
    def to_json(entity):
        return json.dumps(entity)


    @staticmethod
    def from_json(string):
        return json.loads(string)


    @staticmethod
    def decode_map_as_json(dct):
        """
        Used in the __convert_to_array to get an equation from the UI corresponding string.
        """
        for key, value in dct.items():
            if isinstance(value, unicode) and '__mapped_module' in value:
                dict_value = json.loads(value)
                if '__mapped_module' not in dict_value:
                    dct[key] = MapAsJson.decode_map_as_json(dict_value)
                else:
                    modulename = dict_value['__mapped_module']
                    classname = dict_value['__mapped_class']
                    module_entity = __import__(modulename, globals(), locals(), [classname])
                    class_entity = eval('module_entity.' + classname)
                    loaded_entity = class_entity.from_json(value)
                    dct[key] = loaded_entity
        return dct

    class MapAsJsonEncoder(json.JSONEncoder):
        """
        Used before any save to the database to encode Equation type objects.
        """


        def default(self, obj):
            if isinstance(obj, MapAsJson):
                return obj.to_json(obj)
            else:
                return json.JSONEncoder.default(self, obj)


class Sequence(MapAsJson, String):
    """
    Traits type base class that wraps python sequence 
    python types (containers)
    """
    wraps = (dict, list, tuple, set, slice, numpy.ndarray)


class Enumerate(Sequence):
    """
    Traits type that mimics an enumeration.
    """
    wraps = numpy.ndarray


    def __get__(self, inst, cls):
        if inst is None:
            return self
        if self.trait.bound:
            return numpy.array(super(Enumerate, self).__get__(inst, cls))
        return numpy.array(self.trait.value)


    def __set__(self, inst, value):
        if not isinstance(value, list):
            # So it works for simple selects aswell as multiple selects
            value = [value]
        if self.trait.select_multiple:
            super(Enumerate, self).__set__(inst, value)
        else:
            # Bypass default since that only accepts arrays for multiple selects
            setattr(inst, '_' + self.trait.name, self.to_json(value))
            self.trait.value = value


class Dict(Sequence):
    """
    Traits type that wraps a python dict.
    """
    wraps = dict


class Set(Sequence):
    """
    Traits type that wraps a python set.
    """
    wraps = set


class Tuple(Sequence):
    """
    Traits type that wraps a python tuple.
    """
    wraps = tuple


    def __get__(self, inst, cls):
        list_value = super(Tuple, self).__get__(inst, cls)

        if isinstance(list_value, list):
            return list_value[0], list_value[1]

        return list_value


class List(Sequence):
    """
    Traits type that wraps a Python list.
    """
    wraps = (list, numpy.ndarray)


class Slice(Sequence):
    """
    Useful of for specifying views or slices of containers.
    """
    wraps = slice


class Range(core.Type):
    """
    Range is a range that can have multiplicative or additive step sizes.
    Values generated by Range are by default in the interval [start, stop).
    Different flags/modes can be set to include/exclude one or both bounds.
    See the corresponding unittest for more examples.

    Instances of Range will not generate their discrete values automatically,
    but these values can be obtained by converting to a list

    Multiplicative ranges are not yet supported by the web ui.

        >>> range_values = list(Range(lo=0.0, hi=1.0, step=0.1))

        [0.0,
         0.1,
         0.2,
         0.30000000000000004,
         0.4,
         0.5,
         0.6000000000000001,
         0.7000000000000001,
         0.8,
         0.9]

        or by direct iteration:

        >>> for val in Range(lo=0.0, hi=1.0, step=0.1):
                print val

        0.1
        0.2
        0.3
        0.4
        0.5
        0.6
        0.7
        0.8
        0.9


        >>> for val in Range(lo=0.0, hi=2.0, step=1.0):
                print val
        0.0
        1.0
        2.0

        >>> for val in Range(lo=0.0, hi=3.0, step=1.0):
                print val
        0.0
        1.0
        2.0

    using a fixed multiplier
        >>> for val in Range(lo=0.0, hi=5.0, base=2.0):
                print val
        0.0
        2.0
        4.0

        >>> for val in Range(lo=1.0, hi=8.O, base=2.0):
                print val
        1.0
        2.0
        4.0

    """

    # TODO: Improve Range with a fix multiplier and logarithmic range.
    lo = Float(default=0, doc='start of range')
    hi = Float(default=None, doc='end of range')

    step = Float(default=None, doc='fixed step size between elements (for additive). It has priority over "base"')
    base = Float(default=2.0, doc='fixed multiplier between elements')

    # These modes/flags only work for additive step
    MODE_EXCLUDE_BOTH = 0  # flag to exclude both lower and upper bounds
    MODE_INCLUDE_START = 1  # flag to include lo, exclude hi
    MODE_INCLUDE_END = 2  # flag to exclude lo, include hi
    MODE_INCLUDE_BOTH = 3  # flag to include both lower and upper bounds
    mode = Integer(default=1, doc='default behaviour, equivalent to include lo, exclude hi')


    def __iter__(self):
        """ Get valid values in interval"""

        def gen():
            if self.step:
                start, stop, step = self.args_to_decimal(self.lo, self.hi, self.step)
                current = start
                if not self.mode & self.MODE_INCLUDE_START:
                    current += step

                while True:
                    if self.out_of_range(current, stop, self.mode, step):
                        raise StopIteration
                    yield float(current)
                    current += step
            else:
                if self.base <= 1.0:
                    msg = "Invalid base value: %s"
                    LOG.error(msg % str(self.base))
                else:
                    val = self.lo
                    if val == 0:
                        yield val
                        val += self.base
                    while val < self.hi:
                        yield val
                        val *= self.base

        return gen()

    def out_of_range(self, current, stop, mode, step):
        if mode & self.MODE_INCLUDE_END and step > 0:
            return current > stop
        if mode & self.MODE_INCLUDE_END and step < 0:
            return current < stop
        elif step < 0:
            return current <= stop
        return current >= stop


    def args_to_decimal(self, start, stop, step):
        if start > stop and stop is not None:
            step = -step
        if stop is None:
            stop = Decimal(str(start))
            start = Decimal(str(0.0))
        else:
            stop = Decimal(str(stop))
            start = Decimal(str(start))
        step = Decimal(str(step))
        return start, stop, step


class ValidationRange(core.Type):
    """
    ValidationRange represents a Range used only for validating a number.
    """


class JSONType(String):
    """
    Wrapper over a String which holds a serializable object.
    On set/get JSON load/dump will be called.
    """


    def __get__(self, inst, cls):
        if inst:
            string = super(JSONType, self).__get__(inst, cls)
            if string is None or (not isinstance(string, (str, unicode))):
                return string
            if len(string) < 1:
                return None
            return json.loads(string)
        return super(JSONType, self).__get__(inst, cls)


    def __set__(self, inst, value):
        if not isinstance(value, (str, unicode)):
            value = json.dumps(value)
        super(JSONType, self).__set__(inst, value)


class DType(String):
    """
    Traits type that wraps a Numpy dType specification.
    """

    wraps = (numpy.dtype, str)
    defaults = ((numpy.float64,), {})


    def __get__(self, inst, cls):
        if inst:
            type_ = super(DType, self).__get__(inst, cls)
            return str(type_).replace("<type '", '').replace("'>", '')
        return super(DType, self).__get__(inst, cls)


    def __set__(self, inst, value):
        super(DType, self).__set__(inst, str(value))



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
DataTypes for mapping some TVB DataTypes to a Connectivity (regions).

In FreeSurfer terms, a RegionMapping is a parcellation and a VolumeMapping is a segmentation.

.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Mihai Andrei <mihai.andrei@codemart.ro>
"""

import numpy
from tvb.basic.traits import exceptions
from tvb.basic.readers import try_get_absolute_path, FileReader
import tvb.datatypes.arrays as arrays
from tvb.datatypes.connectivity import Connectivity
from tvb.datatypes.surfaces import Surface
from tvb.datatypes.volumes import Volume
from tvb.basic.logger.builder import get_logger
from tvb.basic.arguments_serialisation import parse_slice, preprocess_space_parameters
from tvb.datatypes.structural import VolumetricDataMixin


LOG = get_logger(__name__)


class RegionMapping(arrays.MappedArray):
    """
    An array (of length Surface.vertices). Each value is representing the index in Connectivity regions
    to which the current vertex is mapped.
    """

    array_data = arrays.IndexArray()

    connectivity = Connectivity

    surface = Surface

    __generate_table__ = True

    @staticmethod
    def from_file(source_file="regionMapping_16k_76.txt", instance=None):

        if instance is None:
            result = RegionMapping()
        else:
            result = instance

        source_full_path = try_get_absolute_path("tvb_data.regionMapping", source_file)
        reader = FileReader(source_full_path)

        result.array_data = reader.read_array(dtype=numpy.int32)
        return result

    def get_region_mapping_slice(self, start_idx, end_idx):
        """
        Get a slice of the region mapping as used by the region viewers.
        For each vertex on the surface, alpha-indices will be the closest
        region-index
        :param start_idx: vertex index on the surface
        :param end_idx: vertex index on the surface
        :return: NumPy array with [closest_reg_idx ...]
        """
        if isinstance(start_idx, (str, unicode)):
            start_idx = int(start_idx)
        if isinstance(end_idx, (str, unicode)):
            end_idx = int(end_idx)

        return self.array_data[start_idx: end_idx].T


    def get_triangles_mapping(self):
        """
        :return Numpy array of length triangles and for each the region corresponding to one of its vertices.
        """
        triangles_no = self.surface.number_of_triangles
        result = []
        for i in range(triangles_no):
            result.append(self.array_data[self.surface.triangles[i][0]])
        return numpy.array(result)


    def generate_new_region_mapping(self, connectivity_gid, storage_path):
        """
        Generate a new region mapping with the given connectivity gid from an
        existing mapping corresponding to the parent connectivity.
        """
        new_region_map = self.__class__()
        new_region_map.storage_path = storage_path
        new_region_map._connectivity = connectivity_gid
        new_region_map._surface = self._surface
        new_region_map.array_data = self.array_data
        return new_region_map

    def _find_summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        summary = super(RegionMapping, self)._find_summary_info()
        summary.update({"Source Surface": self.surface.display_name,
                        "Source Surface GID": self.surface.gid,
                        "Connectivity GID": self.connectivity.gid,
                        "Connectivity": self.connectivity.display_name})
        return summary


class RegionVolumeMapping(VolumetricDataMixin, arrays.MappedArray):
    """
    Each value is representing the index in Connectivity regions to which the current voxel is mapped.
    """

    array_data = arrays.IndexArray()

    connectivity = Connectivity

    volume = Volume

    __generate_table__ = True

    apply_corrections = True
    mappings_file = None

    def write_data_slice(self, data):
        """
        We are using here the same signature as in TS, just to allow easier parsing code.
        This method will also validate the data range nd convert it to int, along with writing it is H5.

        :param data: 3D int array
        """

        LOG.info("Writing RegionVolumeMapping with min=%d, mix=%d" % (data.min(), data.max()))
        if self.apply_corrections:
            data = numpy.array(data, dtype=numpy.int32)
            data[data >= self.connectivity.number_of_regions] = -1
            data[data < -1] = -1
            LOG.debug("After corrections: RegionVolumeMapping min=%d, mix=%d" % (data.min(), data.max()))

        if self.mappings_file:
            try:
                mapping_data = numpy.loadtxt(self.mappings_file, dtype=numpy.str, usecols=(0, 2))
                mapping_data = {int(row[0]): int(row[1]) for row in mapping_data}
            except Exception:
                raise exceptions.ValidationException("Invalid Mapping File. Expected 3 columns (int, string, int)")

            if len(data.shape) != 3:
                raise exceptions.ValidationException('Invalid RVM data. Expected 3D.')

            not_matched = set()
            for i in range(data.shape[0]):
                for j in range(data.shape[1]):
                    for k in range(data.shape[2]):
                        val = data[i][j][k]
                        if not mapping_data.has_key(val):
                            not_matched.add(val)
                        data[i][j][k] = mapping_data.get(val, -1)

            LOG.info("Imported RM with values in interval [%d - %d]" % (data.min(), data.max()))
            if not_matched:
                LOG.warn("Not matched regions will be considered background: %s" % not_matched)

        if data.min() < -1 or data.max() >= self.connectivity.number_of_regions:
            raise exceptions.ValidationException("Invalid Mapping array: [%d ... %d]" % (data.min(), data.max()))

        self.store_data("array_data", data)


    def get_voxel_region(self, x_plane, y_plane, z_plane):
        x_plane, y_plane, z_plane = preprocess_space_parameters(x_plane, y_plane, z_plane, self.length_1d,
                                                                self.length_2d, self.length_3d)
        slices = slice(x_plane, x_plane + 1), slice(y_plane, y_plane + 1), slice(z_plane, z_plane + 1)
        voxel = self.read_data_slice(slices)[0, 0, 0]
        if voxel != -1:
            return self.connectivity.region_labels[int(voxel)]
        else:
            return 'background'


    def get_mapped_array_volume_view(self, mapped_array, x_plane, y_plane, z_plane, mapped_array_slice=None, **kwargs):
        x_plane, y_plane, z_plane = preprocess_space_parameters(x_plane, y_plane, z_plane, self.length_1d,
                                                                self.length_2d, self.length_3d)
        slice_x, slice_y, slice_z = self.get_volume_slice(x_plane, y_plane, z_plane)

        if mapped_array_slice:
            matrix_slice = parse_slice(mapped_array_slice)
            measure = mapped_array.get_data('array_data', matrix_slice)
        else:
            measure = mapped_array.get_data('array_data')

        if measure.shape != (self.connectivity.number_of_regions, ):
            raise ValueError('cannot project measure on the space')

        result_x = measure[slice_x]
        result_y = measure[slice_y]
        result_z = measure[slice_z]
        # Voxels outside the brain are -1. The indexing above is incorrect for those voxels as it
        # associates the values of the last region measure[-1] to them.
        # Here we replace those values with an out of scale value.
        result_x[slice_x==-1] = measure.min() - 1
        result_y[slice_y==-1] = measure.min() - 1
        result_z[slice_z==-1] = measure.min() - 1

        return [[result_x.tolist()],
                [result_y.tolist()],
                [result_z.tolist()]]

    def _find_summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        summary = super(RegionVolumeMapping, self)._find_summary_info()
        summary.update({"Source Volume": self.volume.display_name,
                        "Source Volume GID": self.volume.gid,
                        "Connectivity GID": self.connectivity.gid,
                        "Connectivity": self.connectivity.display_name})
        return summary